import lyrebird
import os
import json
import codecs

storage = lyrebird.get_plugin_storage()
CONFIG_FILE = os.path.abspath(os.path.join(storage, 'conf.json'))
DEFAULT_CONF_FILE = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', './default_conf/conf.json'))


class Config:

    def __init__(self):
        self.base_ssh = None
        self.base_path = None


def load():
    if not os.path.exists(CONFIG_FILE):
        f_from = codecs.open(DEFAULT_CONF_FILE, 'r', 'utf-8')
        f_to = codecs.open(CONFIG_FILE, 'w', 'utf-8')
        f_to.write(f_from.read())
        f_to.close()
        f_from.close()
    conf_data = json.loads(codecs.open(CONFIG_FILE, 'r', 'utf-8').read())
    conf = Config()
    conf.__dict__ = conf_data
    return conf



from urllib import parse


# TODO 校验URL是否相等 需要把{num}情况一起放在这里解决

def compare_query(url_org, url_compare):
    param_org = parse.parse_qs(parse.urlparse(url_org).query, True)
    param_compare = parse.parse_qs(parse.urlparse(url_compare).query, True)
    for key, value in param_org.items():
        if not (key in param_compare):
            return False
        if value[0].strip():
            if not param_compare[key] == value:
                return False
    return True




import codecs
import json
import os
import lyrebird
from lyrebird.mock import context
from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.client.jsonscheme import check_filter_schema

CURRENT_DIR = os.path.dirname(__file__)
PLUGIN_DIR = lyrebird.get_plugin_storage()
FILTER_CONF = os.path.join(PLUGIN_DIR, 'filter_conf.json')


'''
过滤处理器
'''

class FilterHandler:
    def init_filter_conf(self):
        if not os.path.exists(FILTER_CONF):
            self.copy_conf_file(FILTER_CONF)

    def copy_conf_file(self, target_path):
        os.path.abspath(os.path.join(CURRENT_DIR, '..', './default_conf/filter_config.json'))
        f_from = codecs.open(os.path.abspath(os.path.join(CURRENT_DIR, '..', './default_conf/filter_config.json')), 'r',
                             'utf-8')
        f_to = codecs.open(target_path, 'w', 'utf-8')
        f_to.write(f_from.read())
        f_to.close()
        f_from.close()

    def get_filer_conf(self):
        self.init_filter_conf()
        try:
            json_obj = json.loads(codecs.open(FILTER_CONF, 'r', 'utf-8').read())
            check_filter_schema(json_obj)
            app_context.filter_dic = json_obj
            msg = json_obj
        except Exception as e:
            msg = '过滤请求的配置文件格式有误:' + e.__getattribute__('message')
        return msg

    def save_filer_conf(self, conf_obj):
        self.init_filter_conf()
        try:
            check_filter_schema(conf_obj)
            f = codecs.open(FILTER_CONF, 'w', 'utf-8')
            f.write(json.dumps(conf_obj))
            f.close()
            # 配置保存后当时生效
            app_context.filter_dic = conf_obj
            return context.make_ok_response()
        except Exception as e:
            msg = '过滤请求的配置文件格式有误:' + e.__getattribute__('message')
            lyrebird.publish('api_coverage', 'error', name='set_filter')
            return context.make_fail_response(msg)



import collections

from jsonschema import validate

schema = {
    "type": "object",
    "properties": {
        "business": {
            "type": "string",
        },
        "version_code": {
            "type": "integer",
        },
        "version_name": {
            "type": "string",
        },
        "api_list": {
            "type": "array",
            "uniqueItems": True,
            "items": {
                "type": "object",
                "properties": {
                    "url": {"description": "The unique identifier for a product", "type": "string", },
                    "desc": {"description": "Name of the product", "type": "string"},
                    "priority": {"type": "integer", }
                },
                "required": ["url", "desc", "priority"]
            }
        }
    },
    "required": ["business", "version_code", "version_name", "api_list"]
}

result_schema = {
    "type": "object",
    "properties": {
        "file_name": {"type": "string"},
        "base_sha1": {"type": "string"},
        "coverage": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "total": {"type": "number"},
                "len": {"type": "integer"},
                "test_len": {"type": "integer"},
                "priorities": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {
                            "type": "object",
                            "properties": {
                                "label": {"type": "integer"},
                                "value": {"type": "number"},
                                "len": {"type": "integer"},
                                "test_len": {"type": "integer"},
                            },
                            "required": ["label", "value", "len", "test_len"]
                    }
                }
            },
            "required": ["name", "total", "len", "test_len","priorities"]
        },
        "test_data": {
            "type": "array",
            "uniqueItems": True,
            "items": {
                "type": "object",
                "properties": {
                    "url": {"description": "The unique identifier for a product", "type": "string", },
                    "desc": {"description": "Name of the product", "type": "string"},
                    "priority": {"type": ["integer", "null"]},
                    "count": {"type": "integer", },
                    "status": {"type": "integer", },
                    "id": {"type": "string"}
                },
                "required": ["url", "desc", "priority"]
            }
        }
    },
    "required": ["file_name", "base_sha1", "coverage", "test_data"]
}

filter_schema = {
    "type": "object",
    "properties": {
        "exclude": {
            "type": "object",
            "properties": {
                "host": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string", },
                },
                "regular": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string", }
                },
            },
            "required": ["host", "regular"]
        }
    },
    "required": ["exclude"]
}


def check_url_redundant(obj):
    repeat_urls = []
    url_list = list(map(lambda x: x.get('url'), obj.get('api_list')))
    url_count = dict(collections.Counter(url_list))
    for k, v in url_count.items():
        if v > 1:
            repeat_urls.append(k)
    return repeat_urls


def check_schema(obj):
    validate(obj, schema)


def check_filter_schema(obj):
    validate(obj, filter_schema)


def check_result_schema(obj):
    validate(obj, result_schema)


"""
事件处理器
接收用户信息和设备信息等
"""

import lyrebird
from lyrebird_api_coverage.client.context import app_context


def event_handler(event):
    app_context.info = event.get('message')


def user_handler(event):
    app_context.user_info = event.get('message')


def event_subscribe():
    lyrebird.subscribe('device_info', event_handler)
    lyrebird.subscribe('user_info', user_handler)



import re
from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.handlers.filter_handler import FilterHandler

"""
过滤器，滤除不关注的API
"""
class Filter:
    def filter_host(self, url, host_list):
        if url.split('/')[0] in host_list:
            return True
        else:
            return False

    def filter_re(self, url, pattern_list):
        flag = False
        for pattern_item in pattern_list:
            if re.search(pattern_item, url) is not None:
                return True
            else:
                flag = False
        return flag

    def filter_all(self, url):
        # 如果没有init filter conf 做个处理
        if not app_context.filter_dic:
            FilterHandler().get_filer_conf()
        host_list = app_context.filter_dic.get('exclude').get('host')
        pattern_list = app_context.filter_dic.get('exclude').get('regular')
        host_filter = self.filter_host(url, host_list)
        regular_filter = self.filter_re(url, pattern_list)
        if host_filter or regular_filter is True:
            return True
        else:
            return False



import codecs
import json
import os
import time

import lyrebird

from lyrebird_api_coverage.client.context import app_context

from lyrebird_api_coverage.client.merge_algorithm import mergeAlgorithm

CURRENT_DIR = os.path.dirname(__file__)

PLUGINS_CONF_DIR = lyrebird.get_plugin_storage()
PLUGINS_DATA_DIR = os.path.join(PLUGINS_CONF_DIR, 'data')
PLUGINS_DUMP_DIR = os.path.join(PLUGINS_CONF_DIR, 'dump')

if not os.path.exists(PLUGINS_DATA_DIR):
    os.makedirs(PLUGINS_DATA_DIR)

if not os.path.exists(PLUGINS_DUMP_DIR):
    os.makedirs(PLUGINS_DUMP_DIR)

"""
ResultHandler 测试结果处理器
保存测试结果、续测、清空测试结果

"""

class ResultHandler:

    def make_result_dir(self):
        save_time = time.strftime('%Y-%m-%d-%H:%M:%S', time.localtime(time.time()))
        test_data_dir = os.path.join(PLUGINS_DATA_DIR, str(save_time))
        if not os.path.exists(test_data_dir):
            os.makedirs(test_data_dir)
        return test_data_dir

    def save_result(self, filename):
        # test_data_dir = self.make_result_dir()
        test_data_dir = PLUGINS_DATA_DIR
        test_path = os.path.join(test_data_dir, str(filename) + '.json')
        f = codecs.open(test_path, 'w', 'utf-8')
        result_json = json.dumps({'business': app_context.business, 'verison_name': app_context.version_name,
                                  'verison_code': app_context.version_code, 'file_name': app_context.filename,
                                  'base_sha1': app_context.base_sha1, 'coverage': app_context.coverage,
                                  'test_data': app_context.merge_list}, ensure_ascii=False, indent=4)
        f.write(result_json)
        f.close()

    def dump_info(self, filename):
        # test_data_dir = self.make_result_dir()
        dump_dir = PLUGINS_DUMP_DIR
        test_path = os.path.join(dump_dir, str(filename) + '.json')
        f = codecs.open(test_path, 'w', 'utf-8')
        result_json = json.dumps({'business': app_context.business, 'verison_name': app_context.version_name,
                                  'verison_code': app_context.version_code, 'file_name': app_context.filename,
                                  'base_sha1': app_context.base_sha1, 'coverage': app_context.coverage,
                                  'test_data': app_context.merge_list}, ensure_ascii=False, indent=4)
        f.write(result_json)
        f.close()

    def resume_test(self):
        test_path = os.path.join(PLUGINS_DATA_DIR, 'result.json')
        if not os.path.exists(test_path):
            dic = {'status': 'error'}
        else:
            json_obj = json.loads(codecs.open(test_path, 'r', 'utf-8').read())
            app_context.merge_list = json_obj.get('test_data')
            mergeAlgorithm.coverage_handler()
            dic = {'status': 'success',
                   'message': {'coverage': app_context.coverage, 'test_data': app_context.merge_list}}
        return dic

    def clear_cache_result(self):
        # 清空
        app_context.user_list.clear()
        app_context.base_list.clear()
        app_context.user_org_list.clear()
        # 清空coverage
        app_context.coverage.clear()
        # 清空缓存list
        app_context.merge_list.clear()
        # 清空sha1
        app_context.base_sha1 = None



import lyrebird
from lyrebird_api_coverage.client import format_url
from lyrebird_api_coverage.client.context import app_context
import time


class MergeAlgorithm:

    def first_result_handler(self, json_obj):
        """
        初始化/首次/切换base之后 需要重新生成base_list，merge_list
        注意没有初始化coverage
        coverage初始化需要在getCoverage接口实现
        """

        # 清空
        app_context.user_list.clear()
        app_context.base_list.clear()
        app_context.user_org_list.clear()
        # 清空coverage
        app_context.coverage.clear()
        # 清空缓存list
        app_context.merge_list.clear()
        # 清空 参数列表中间处理数据
        app_context.path_param_dic.clear()
        # 初始化base数据，format成需要的数据格式
        self.init_basedata_handler(json_obj)
        # 初始化coverage数据
        self.coverage_arithmetic(json_obj)
        # 获取所有的base URL
        app_context.base_list = list(map(lambda x: x.get('url'), json_obj.get('api_list')))

    def init_basedata_handler(self, dic):
        # for k, v in dic.items():
        #     url_dic = {'url': k, 'desc': v.get('desc'), 'priority': v.get('priority'), 'count': 0, 'status': 0,
        #                'org': []}
        #     app_context.merge_list.append(url_dic)
        dict2 = {'count': 0, 'status': 0, 'id': ''}
        for item in dic.get('api_list'):
            # 处理带参数的情况
            if '?' in item['url']:
                path = item['url'].split('?')[0].lower()
                params = item['url'].split('?')[1].split('&')
                param_dic = {}
                for i in params:
                    key = i.split('=')[0]
                    val = i.split('=')[1]
                    param_dic[key] = val

                if app_context.path_param_dic.get(path):
                    app_context.path_param_dic[path].append({'url': item['url'], 'params': param_dic,
                                                             'url_base': format_url.format_api_source(
                                                                 item.get('url')).lower()})
                else:
                    app_context.path_param_dic[path] = [{'url': item['url'], 'params': param_dic,
                                                         'url_base': format_url.format_api_source(
                                                             item.get('url')).lower()}]

            # format base源 同时变成大小写归一化，变小写
            item['url'] = format_url.format_api_source(item.get('url')).lower()
            item.update(dict2)
            app_context.merge_list.append(item)

    def merge_handler_new(self, user_url, path_id):
        """
        status=0 base中包含未覆盖，status=1 base中包含已覆盖，status=2 base中不包含且覆盖到的;
        path_id表示URL的handler_context的唯一标识，查看详情用
        """

        # 在list中筛选出想要的数据,筛选结果直接取0即可
        specific_filter_list = list(filter(lambda x: x.get('url') == user_url, app_context.merge_list))

        # 判断筛选出来的list是否为空,即是否在list中存在
        if specific_filter_list:
            specific_dic = specific_filter_list[0]
            # 移除掉对应的数据为插入index0的位置做前置处理
            app_context.merge_list.remove(specific_dic)
            # 做业务处理
            if specific_dic['status'] == 0:
                specific_dic['status'] = 1
                # 把首次覆盖到的API,放入user_list里面
                app_context.user_list.append(user_url)
            # count +1
            specific_dic['count'] += 1  # 插入原始url  # specific_dic['org'].append(org_url)
            specific_dic['id'] = path_id
        else:
            # specific_dic = {'url': user_url, 'desc': '', 'priority': '', 'count': 1, 'status': 2, 'org': [org_url]}
            specific_dic = {'url': user_url, 'desc': '', 'priority': None, 'count': 1, 'status': 2, 'id': path_id}
        # 插入到 index=0 的位置
        app_context.merge_list.insert(0, specific_dic)

    def coverage_handler(self):
        """
        总体覆盖率
        """
        # 获取handle前的历史覆盖率为做对比用
        history_coverage = app_context.coverage['total']
        test_len = len(list(filter(lambda x: x.get('status') == 1, app_context.merge_list)))
        if app_context.coverage['len'] == 0:
            coverage = 0
        else:
            coverage = round(test_len / app_context.coverage['len'] * 100, 2)
        # 为了传给Overbridge的socket信息format数据格式
        app_context.coverage['total'] = coverage
        # 覆盖率有变化才emit & publish 覆盖率的变化消息给API-Coverage前端，overbridge前端，和消息总线
        if not history_coverage == coverage:
            handler_time = time.time()
            # 限制频繁emit io msg，在两次之间大于指定时间间隔才会emit
            if handler_time - app_context.covtime > app_context.SOCKET_PUSH_INTERVAL:
                lyrebird.emit('coverage message', app_context.coverage.get('total'), namespace='/api_coverage')
                app_context.covtime = handler_time
            by_priority = [p.get('value') for p in app_context.coverage['priorities']]
            lyrebird.publish('coverage', 
                dict(
                    name='coverage', 
                    value=app_context.coverage.get('total'), 
                    by_priority=by_priority)
            )
        app_context.coverage['test_len'] = test_len
        # 各优先级对应覆盖率
        for item_dic in app_context.coverage.get('priorities'):
            item_length = item_dic.get('len')
            test_item_length = len(list(
                filter(lambda x: x.get('priority') == item_dic.get('label') and x.get('status') == 1,
                       app_context.merge_list)))
            if item_length == 0:
                coverage = 0
            else:
                coverage = round(test_item_length / item_length * 100, 2)
            item_dic['value'] = coverage
            item_dic['test_len'] = test_item_length

    def init_resume_data(self, dic):
        app_context.user_list = []
        app_context.base_list = []
        for k, v in dic.items():
            if v['status'] == 1:
                app_context.user_list.append(k)
                app_context.base_list.append(k)
            elif v['status'] == 0:
                app_context.base_list.append(k)

    # 获取优先级初始info，优先级分几个，以及各个优先级的list长度,传入的也是conf文件
    def coverage_arithmetic(self, dic):
        url_info_list = dic.get('api_list')
        # 获取优先级list，非空
        priority_list = list(set(list(map(lambda x: x.get('priority'), url_info_list))))
        # 去除空值
        if list(filter(lambda x: x == '', priority_list)):
            priority_list.remove('')
        # 排序
        app_context.priority_list = sorted(priority_list)
        # init coverage 原始数据结构 total总体数据
        app_context.coverage['name'] = 'coverage'
        app_context.coverage['total'] = 0
        app_context.coverage['len'] = len(url_info_list)
        app_context.coverage['test_len'] = 0
        app_context.coverage['priorities'] = []
        # 各个优先级的init数据
        for item in app_context.priority_list:
            item_length = len(list(filter(lambda x: x.get('priority') == item, url_info_list)))
            coverage = 0
            app_context.coverage['priorities'].append(
                {'label': item, 'value': coverage, 'len': item_length, 'test_len': 0})

    # 对import的打算resume的result和 缓存里面的测试结果进行merge
    def merge_resume(self, result_list):
        # 在list中筛选出url,筛选结果直接取0即可
        cache_list = app_context.merge_list
        # 取出url信息，组成list
        cache_url_list = list(map(lambda x: x.get('url'), cache_list))
        result_url_list = list(map(lambda x: x.get('url'), result_list))
        # 找到交集
        intersection_list = list(set(cache_url_list).intersection(set(result_url_list)))
        # 对交集进行处理
        for url in intersection_list:
            cache_spec = list(filter(lambda x: x.get('url') == url, cache_list))[0]
            result_spec = list(filter(lambda x: x.get('url') == url, result_list))[0]
            spec_dict = list(filter(lambda x: x.get('url') == url, app_context.merge_list))[0]
            app_context.merge_list.remove(spec_dict)
            # 修改count status org_url
            spec_dict['count'] = cache_spec.get('count') + result_spec.get('count')
            if spec_dict['status'] == 0 and spec_dict['count'] != 0:
                spec_dict['status'] = 1
            app_context.merge_list.insert(0, spec_dict)
        # 找到差集
        diff_url_list = list(set(result_url_list) - set(cache_url_list))
        # 差集处理（对result中存在，但cache不存在的）status=2的情景进行extend处理
        diff_list = list(filter(lambda x: True if x.get('url') in diff_url_list else False, result_list))
        app_context.merge_list.extend(diff_list)
        # 重新计算覆盖率
        self.coverage_handler()


mergeAlgorithm = MergeAlgorithm()



import codecs
import hashlib
import json

import lyrebird
from flask import request
from lyrebird_api_coverage.client.load_base import DEFAULT_BASE

from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.client.merge_algorithm import mergeAlgorithm
from lyrebird_api_coverage.handlers.base_source_handler import DEFAULT_BASE, BaseDataHandler

from lyrebird import context

"""
ImportHandler

base导入处理器
result导入处理器

"""

class ImportHandler:

    def import_base_handler(self):
        json_file = request.files.get('json-import')
        mimetype = json_file.content_type
        # 判断是不是json格式的文件
        if mimetype == 'application/json':
            # 读取文件流，注意文件流只能read一次
            bytes_obj = json_file.read()
            try:
                check_result = BaseDataHandler().check_base(json.loads(bytes_obj))
                if check_result:
                    return check_result

                self.write_wb(DEFAULT_BASE, bytes_obj)
                # 读取json文件
                json_obj = json.loads(codecs.open(DEFAULT_BASE, 'r', 'utf-8').read())
                # 获取文件的sha1
                app_context.base_sha1 = self.get_sha1(bytes_obj)
                # 初次处理，切换后的result
                mergeAlgorithm.first_result_handler(json_obj)
                mergeAlgorithm.coverage_arithmetic(json_obj)
                resp = context.make_ok_response()
                lyrebird.publish('api_coverage', 'operation', name='import_base')
            except Exception as e:
                resp = context.make_fail_response('导入文件内容格式有误:' + str(e))
                lyrebird.publish('api_coverage', 'error', name='import_base')
        else:
            resp = context.make_fail_response("Error.The selected non - JSON file.")
            lyrebird.publish('api_coverage', 'error', name='import_base')
        return resp

    def import_result_handler(self):
        json_file = request.files.get('json-import')
        mimetype = json_file.content_type
        # 读取文件流，注意文件流只能read一次
        bytes_obj = json_file.read()
        try:
            result_obj = json.loads(bytes_obj)
            # 获取import文件的sha1
            import_sha1 = result_obj.get('base_sha1')
            if app_context.base_sha1 == import_sha1:
                # merge import result and cache result
                # check_result_schema(result_obj)
                app_context.coverage = json.loads(bytes_obj).get('coverage')
                mergeAlgorithm.merge_resume(result_obj.get('test_data'))
                # 放入缓存
                # app_context.merge_list = json.loads(bytes_obj).get('test_data')
                # app_context.coverage = json.loads(bytes_obj).get('coverage')
                resp = context.make_ok_response()
                lyrebird.publish('api_coverage', 'operation', name='import_result')
            else:
                resp = context.make_fail_response('导入的测试结果和之前选择base不匹配')
                lyrebird.publish('api_coverage', 'error', name='import_result')
        except Exception as e:
            resp = context.make_fail_response('导入文件内容格式有误:' + str(e))
            lyrebird.publish('api_coverage', 'error', name='import_result')
        return resp

    def write_wb(self, path, obj):
        f = codecs.open(path, 'wb')
        f.write(obj)
        f.close()

    def get_sha1(self, obj):
        sha1obj = hashlib.sha1()
        sha1obj.update(obj)
        hash_sha1 = sha1obj.hexdigest()
        return hash_sha1



from lyrebird.plugins import manifest
from . import api
from . import interceptor
from flask import Response
from lyrebird_api_coverage.handlers.base_source_handler import BaseDataHandler
from lyrebird_api_coverage.client.merge_algorithm import mergeAlgorithm
from lyrebird_api_coverage.client.event_subscibe import event_subscribe

# 执行插件初始化操作
# 获取base_data_config文件信息
base_dict = BaseDataHandler().get_base_source()
# 如果import的文件异常
if not isinstance(base_dict, Response):
    mergeAlgorithm.first_result_handler(base_dict)
    mergeAlgorithm.coverage_arithmetic(base_dict)
# 总线消息订阅
event_subscribe()

manifest(
    id='api_coverage',
    name='APICoverage',
    api=[
        # 获取内存里保存的测试结果API
        ('/api/getTest', api.get_test_data),
        # 获取内存里保存的测试覆盖率信息
        ('/api/getCoverage', api.get_coverage),
        # 保存测试数据在本地
        ('/api/saveResult', api.save_result,['POST']),
        # 续传测试结果
        ('/api/resumeTest', api.resume_test, ['POST']),
        # 清空测试缓存结果
        ('/api/clearResult', api.clear_result),
        # 导入base json文件
        ('/api/importBase', api.import_base, ['POST']),
        # 获取filter的conf文件
        ('/api/getFilterConf', api.get_filter_conf),
        # 覆盖配置filter conf文件
        ('/api/setFilterConf', api.set_filter_conf, ['POST']),
        # overbridge dump 信息用的API
        ('/api/dump', api.dump),
        # baseInfo
        ('/api/baseInfo', api.get_base_info),
    ],
    background=[
    ],
    event=[
        ('flow', interceptor.on_request)
    ]
)



import json
import lyrebird
import os
from flask import request, jsonify, Response, stream_with_context
from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.handlers.base_source_handler import BaseDataHandler
from lyrebird_api_coverage.client.merge_algorithm import mergeAlgorithm
from lyrebird_api_coverage.handlers.filter_handler import FilterHandler
from lyrebird_api_coverage.handlers.import_file_handler import ImportHandler
from lyrebird_api_coverage.handlers.result_handler import ResultHandler, PLUGINS_DUMP_DIR
from lyrebird import context

def generate(data):
    rtext = json.dumps(data)
    yield rtext

# 获取内存里保存的测试结果API
# /getTest
def get_test_data():
    return Response(stream_with_context(generate({'test_data': app_context.merge_list})), content_type='application/json')

# 获取内存里保存的测试覆盖率信息
# /getCoverage
def get_coverage():
    return Response(stream_with_context(generate(app_context.coverage)), content_type='application/json')

# 保存测试数据在本地
#  /saveResult
def save_result():
    # 传入文件名
    filename = request.form.get('result_name')
    ResultHandler().save_result(filename)
    lyrebird.publish('api_coverage', 'operation', name='save_result')
    return context.make_ok_response()

#续传测试结果
# /resumeTest
def resume_test():
    # resp = ResultHandler().resume_test()
    resp = ImportHandler().import_result_handler()
    return resp

# 清空测试缓存结果
# /clearResult
def clear_result():
    ResultHandler().clear_cache_result()
    # 获取基准文件
    base_dict = BaseDataHandler().get_base_source()
    # 初始化正常会进行数据的处理：覆盖率初始化 & API LIST初始化
    if not isinstance(base_dict, Response):
        mergeAlgorithm.first_result_handler(base_dict)
        mergeAlgorithm.coverage_arithmetic(base_dict)
    lyrebird.publish('api_coverage', 'operation', name='clear_result')
    return context.make_ok_response()

# 导入base json文件
# /importBase
def import_base():
    resp = ImportHandler().import_base_handler()
    return resp

# 获取filter的conf文件
# /getFilterConf
def get_filter_conf():
    msg = FilterHandler().get_filer_conf()
    # 如果返回的string包含报错信息，则是报错
    if isinstance(msg, str):
        return context.make_fail_response(msg)
    else:
        return jsonify(msg)

# 覆盖配置filter conf文件
# /setFilterConf
def set_filter_conf():
    filter_data = request.form.get('filter_data')
    try:
        resp = FilterHandler().save_filer_conf(json.loads(filter_data))
        lyrebird.publish('api_coverage', 'operation', name='set_filter')
    except Exception as e:
        lyrebird.publish('api_coverage', 'error', name='set_filter')
        return context.make_fail_response("传入的非json文件" + str(repr(e)))
    return resp

# overbridge dump 信息用的API
# /dump
def dump():
    filename = 'api_coverage'
    ResultHandler().dump_info(filename)
    return jsonify(
        [{'name': str(filename) + '.json', 'path': os.path.join(PLUGINS_DUMP_DIR, str(filename) + '.json')}])

# base info
# /baseInfo
def get_base_info():
    return context.make_ok_response(
        business=app_context.business,
        version_name=app_context.version_name,
        version_code=app_context.version_code
        )



import codecs
import hashlib
import json
import lyrebird
from lyrebird.log import get_logger
import os

from lyrebird_api_coverage.client.context import app_context

PLUGINS_CONF_DIR = lyrebird.get_plugin_storage()
DEFAULT_BASE = os.path.join(PLUGINS_CONF_DIR, 'base.json')
CURRENT_DIR = os.path.dirname(__file__)


def get_file_sha1(path):
    with open(path, 'rb') as f:
        sha1obj = hashlib.sha1()
        sha1obj.update(f.read())
        hash_sha1 = sha1obj.hexdigest()
        return hash_sha1


def auto_load_base():
    lyrebird_conf = lyrebird.context.application.conf
    # 读取指定base文件，写入到base.json
    if lyrebird_conf.get('hunter.base'):
        base_path = lyrebird_conf.get('hunter.base')
        base = codecs.open(base_path, 'r', 'utf-8').read()
        f = codecs.open(DEFAULT_BASE, 'w', 'utf-8')
        f.write(base)
        f.close()
        app_context.base_sha1 = get_file_sha1(DEFAULT_BASE)
        return json.loads(base)
    # 通过本地默认base文件获取base
    elif not os.path.exists(DEFAULT_BASE):
        copy_file(DEFAULT_BASE)
    with codecs.open(DEFAULT_BASE, 'r', 'utf-8') as f:
        json_obj = json.load(f)
        app_context.base_sha1 = get_file_sha1(DEFAULT_BASE)
        return json_obj


def copy_file(target_path):
    os.path.abspath(os.path.join(CURRENT_DIR, '..', './default_conf/base.json'))
    f_from = codecs.open(os.path.abspath(os.path.join(CURRENT_DIR, '..', './default_conf/base.json')), 'r', 'utf-8')
    f_to = codecs.open(target_path, 'w', 'utf-8')
    f_to.write(f_from.read())
    f_to.close()
    f_from.close()



import lyrebird
from lyrebird import log
from lyrebird_api_coverage.client import format_url
from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.client.merge_algorithm import mergeAlgorithm
from lyrebird_api_coverage.client.report import report_worker
from lyrebird_api_coverage.client.url_compare import compare_query

logger = log.get_logger()
import time

def on_request(msg):
    req_starttime = time.time()
    req_msg = msg['flow']
    logger.debug(req_msg)
    if not msg['flow']['request']['url']:
        return
    short_url = msg['flow']['request']['url'].replace('http://', '').replace('https://', '').split('?')[0]
    # format之后的真正PATH，处理{num}这样的情况，emit给前端，做刷新table用，同时处理成小写
    path = format_url.format_api(short_url).lower()
    # 获取handler_context.id，为前端展开看详情准备
    path_id = msg['flow']['id']
    device_ip = msg['flow']['client_address']
    if path in app_context.base_list:
        # merge到 context merge list中
        mergeAlgorithm.merge_handler_new(path, path_id)
        # 在base里的需要去计算下覆盖率
        mergeAlgorithm.coverage_handler()
        # 进行上报
        report_worker(path, device_ip)
        # 计算差值，指定时间间隔内只发1次io msg，限制刷新频率
        emit(req_starttime, path)
    # 如果path配置了对应的参数
    elif path in app_context.path_param_dic:
        ulr_list = app_context.path_param_dic[path]
        flag = 0
        for item in ulr_list:
            if compare_query(item['url'], msg['flow']['request']['url']):
                mergeAlgorithm.merge_handler_new(item['url_base'], path_id)
                mergeAlgorithm.coverage_handler()
                report_worker(item['url_base'], device_ip)
                flag = 1
        # 如果参数组合不存在，提取关注的字段
        if flag == 0:
            url_pgroup = ''
            params_list = []
            for item in ulr_list:
                params_list.extend(item['params'].keys())
            # 去重
            for p in list(set(params_list)):
                # Todo 这里在初始化之后看一下
                val = msg['flow']['request']['query'][p]
                if url_pgroup:
                    url_pgroup = url_pgroup + '&' + str(p) + '=' + str(val)
                else:
                    url_pgroup = path + '?' + str(p) + '=' + str(val)
            mergeAlgorithm.merge_handler_new(url_pgroup, path_id)
            mergeAlgorithm.coverage_handler()
            report_worker(url_pgroup, device_ip)
        # 计算差值，指定时间间隔内只发1次io msg，限制刷新频率
        emit(req_starttime, path)
    # 如果不在base里，不需要merge到数据中
    else:
        # mergeAlgorithm.merge_handler_new(path, path_id)
        # 进行上报
        report_worker(path, device_ip)

def emit(starttime, path):
        duration = starttime - app_context.endtime
        if duration > app_context.SOCKET_PUSH_INTERVAL:
            app_context.endtime = starttime
            lyrebird.emit('apiCoverageBaseData')



import re


def format_api(url):
    # 正则匹配例如 11,22,33 或 只有 11 这类的数据，同时排除出现 123test 这样的规则匹配，加上/作为过滤标准，或者以/123结尾的匹配
    # m = re.sub("(\/(\d+,){0,}\d+\/)", "/{num}/", url)
    # api = re.sub("(\/(\d+,){0,}\d+\Z)", "/{num}", m)
    m = re.sub('(\/(0|[1-9][0-9]*|-[1-9][0-9]*)\/)', "/{num}/", url)
    api = re.sub('(\/(0|[1-9][0-9]*|-[1-9][0-9]*))', "/{num}", m)
    return api


def format_api_source(url):
    # 正则匹配例如 apitrip.meituan.com/volga/api/v1/{trip}/home/{aaaa}/z
    short_url = re.sub("(\/\{.*?\})+", "/{num}", url)
    return short_url



import os

import lyrebird
from lyrebird.mock import context

from lyrebird_api_coverage.client.context import app_context
from lyrebird_api_coverage.client.jsonscheme import check_schema, check_url_redundant
from lyrebird_api_coverage.client.load_base import auto_load_base

CURRENT_DIR = os.path.dirname(__file__)

PLUGINS_CONF_DIR = lyrebird.get_plugin_storage()

if not os.path.exists(PLUGINS_CONF_DIR):
    os.makedirs(PLUGINS_CONF_DIR)

# 默认base 文件
DEFAULT_BASE = os.path.join(PLUGINS_CONF_DIR, 'base.json')

'''
Base 处理器
'''
class BaseDataHandler:

    '''
    获取base文件
    '''
    def get_base_source(self):
        json_obj = auto_load_base()
        if not json_obj:
            json_obj = context.make_fail_response('暂无默认文件,需手动导入base文件')
        # 检查不为空的base文件是否符合标准,符合标准check_base返回0
        else:
            if self.check_base(json_obj):
                # 遇到异常就返回
                return self.check_base(json_obj)
        return json_obj

    '''
    检查base是否符合规则
    '''
    def check_base(self, obj):
        try:
            # 检查base schema
            check_schema(obj)
            # 检查url是否有重复项存在
            redundant_items = check_url_redundant(obj)
            if redundant_items:
                resp = context.make_fail_response('导入API有重复项' + str(redundant_items))
                lyrebird.publish('api_coverage', 'error', name='import_base')
                return resp
            # 获取base内容，解析出base的business等字段
            filename = obj.get('business') + obj.get('version_name') + '.' + str(obj.get('version_code'))
            app_context.filename = filename
            app_context.business = obj.get('business')
            app_context.version_name = obj.get('version_name')
            app_context.version_code = obj.get('version_code')
            return
        except Exception as e:
            resp = context.make_fail_response(
                '导入文件有误:' + '\n' + e.__getattribute__('message') + '\n' + '！请重新import base')

        return resp



from lyrebird_api_coverage.client.context import app_context
from lyrebird import report

"""
上报处理器，用户请求行为上报到ELK中
ps:需要在lyrebird中设定reporter相关配置
"""

class ReportHandler:
    def check_url_info(self, url, device_ip):
        specific_list = list(filter(lambda x: x.get('url') == url, app_context.merge_list))
        if specific_list and specific_list[0].get('status') == 1:
            desc = specific_list[0].get('desc')
            count_flag = 1
            priority = specific_list[0].get('priority')
        else:
            desc = 'N/A'
            count_flag = -1
            priority = -1
        info_dict = {'url': url, 'desc': desc, 'priority': priority, 'count_flag': count_flag,
                     'business': app_context.business, 'version_name': app_context.version_name,
                     'version_code': app_context.version_code}
        if app_context.info.get(device_ip):
            # 如果有Device信息，就上报device相关的信息
            info_dict.update(app_context.info.get(device_ip))
        return info_dict


report_handler = ReportHandler()


def report_worker(url, device_ip):
    update_data = report_handler.check_url_info(url, device_ip)
    update_data.update({"action": "api-coverage", "user_info": app_context.user_info})
    report(update_data)



"""
上下文，保存缓存数据
"""
class Context:
    def __init__(self):
        # base url list
        self.base_list = []
        # user tested url list（base包含的）
        self.user_list = []
        # 用户访问的原始url
        self.user_org_list = []
        # 前端需要的list，base数据和user访问数据 merge之后的结果list
        self.merge_list = []
        # 参数校验需要的dict
        self.path_param_dic = {}
        # 优先级级名list
        self.priority_list = []
        # coverage 覆盖率信息
        self.coverage = {}
        # base文件对应的sha1
        self.base_sha1 = ''
        # base文件对应的filename
        self.filename = ''
        # 过滤规则包含host和regular
        self.filter_dic = {}
        # device & APP信息
        self.info = {}
        # 来自base文件的信息
        self.business = ''
        self.version_name = ''
        self.version_code = None
        # user_info
        self.user_info = {}
        # 记录请求最后的时间，避免频繁emit io消息
        self.endtime = 0
        # 记录上次coverage变化的时间，避免频繁emit io消息
        self.covtime = 0
        # 时间间隔，每隔指定时间触发1次socket io消息，防止刷新频繁
        self.SOCKET_PUSH_INTERVAL = 1


# 单例模式
app_context = Context()
